from collections import defaultdict
from lkb.lkb_base_class import Conceptual_Edge, Lexical_Edge, Lexical_Knowledge_Base


class Basic_Lexical_Knowledge_Base(Lexical_Knowledge_Base):#Good for simple queries, not memory-efficient
    def __init__(self,concepts,terms,conceptual_relations,lexical_edges,conceptual_edges):
        super().__init__(concepts,terms,conceptual_relations,lexical_edges,conceptual_edges)
        self.id_to_concept = {concept.id:concept for concept in concepts}
        self.id_to_term= {term.id:term for term in terms}
        self.id_to_concept_relations = {relation.id:relation for relation in conceptual_relations}
        self.concept_relation_name_to_concept_relation = {relation.string:relation for relation in conceptual_relations}
        self.outward_conceptual_edges = defaultdict(list)
        self.inward_conceptual_edges = defaultdict(list)
        self.concept_id_to_term_ids = defaultdict(list)
        self.term_id_to_concept_ids = defaultdict(list)
        print("Creating conceptual edge index")
        for edge in conceptual_edges:
            self.outward_conceptual_edges[edge.concept_id_1].append((self.id_to_concept_relations[edge.rel_id],self.id_to_concept[edge.concept_id_2]))
            self.inward_conceptual_edges[edge.concept_id_2].append((self.id_to_concept_relations[edge.rel_id],self.id_to_concept[edge.concept_id_1]))
        print("Creating lexical edge index")
        for edge in lexical_edges:
            self.concept_id_to_term_ids[edge.concept_id].append(edge.term_id)
            self.term_id_to_concept_ids[edge.term_id].append(edge.concept_id)
    # def contains_concept(self,concept_id):
    #     return concept_id in self.id_to_concept
    def get_concept(self,concept_id:str):
        return self.id_to_concept[concept_id]  
    def get_term(self,term_id):
        return self.id_to_term[term_id]  
    def get_relation(self,rel_id):
        return self.id_to_concept_relations[rel_id]
    def get_outward_edges(self,subject_concept_id:str):
        return self.outward_conceptual_edges[subject_concept_id]
    def get_inward_edges(self,object_concept_id:str):
        return self.inward_conceptual_edges[object_concept_id]    
    def get_terms_from_concept_id(self,concept_id:str):
        return [self.get_term(term_id) for term_id in self.concept_id_to_term_ids[concept_id]]
    def get_concepts_from_term_id(self,term_id:str):
        return [self.get_concept(term_id) for term_id in self.term_id_to_concept_ids[term_id]]
    def get_data(self):
        concepts =  set(self.id_to_concept.values())
        terms = set(self.id_to_term.values())
        conceptual_relations = set(self.id_to_concept_relations.values())
        conceptual_edges = set()
        for concept_id_1, related_concepts in self.outward_conceptual_edges.items():
            for rel,concept_2 in related_concepts:
                conceptual_edges.add(Conceptual_Edge(concept_id_1,concept_2.id,rel.id))
        lexical_edges = set()
        for concept_id,term_ids in self.concept_id_to_term_ids.items():
            for term_id in term_ids:
                lexical_edges.add(Lexical_Edge(concept_id,term_id))
        return concepts,terms,conceptual_edges,lexical_edges,conceptual_relations
    def get_concept_ids(self):
        return set(self.id_to_concept.keys())

    






